# coding: utf-8

from kivy.app import App
from kivy.uix.boxlayout import BoxLayout
from kivy.properties import ObjectProperty

from pysga import sga
from pysga.configure import ParamsSGA, ObjectiveFunc


class SGALayout(BoxLayout):

    resultado = ObjectProperty()
    screen_manager = ObjectProperty()
    get_from_file = ObjectProperty()

    def __init__(self, **kwargs):
        super(SGALayout, self).__init__(**kwargs)
        self.get_from_file.bind(active=self.on_checkbox_active)

    def otimizar(self):
        try:
            x, y = sga.run(sga=self.configure_sga(), fobj=self.configure_function())
            texto = '  Minimum point:\n  %s\n  Optimal value:\n  %s' % (str(x), str(y))
        except:
            texto = '  SGA Error\n  Verify parameters and objective function'
        self.resultado.text = texto

    def configure_sga(self):
        alfamin = float(self.ids['alfamin'].text)
        alfainitial = float(self.ids['alfainitial'].text)
        iteratios = int(self.ids['iterations'].text)
        global_ratio = float(self.ids['global_ratio'].text)
        population = int(self.ids['population'].text)
        search_group_ratio = float(self.ids['search_group_ratio'].text)
        n_pertubed = int(self.ids['n_pertubed'].text)
        tournament = int(self.ids['tournament'].text)
        small_value = float(self.ids['small_value'].text)

        return ParamsSGA(AlfaMin=alfamin, AlfaInitial=alfainitial, NIterations=iteratios,
                        GlobalIterationsRatio=global_ratio, PopulationSize=population,
                        SearchGroupRatio=search_group_ratio, NPerturbed=n_pertubed,
                        TournamentSize=tournament, SmallValue=small_value)

    def configure_function(self):
        lowers = self.get_list('lowers')
        uppers = self.get_list('uppers')
        if self.get_from_file.active:
            try:
                from fobj_function import fobj
            except ImportError:
                from pysga.fobj_function import fobj
            fobj_func = fobj
        else:
            fobj_func = self.get_fobj_function()
        fobj = ObjectiveFunc(inf=lowers, sup=uppers)
        fobj.evaluate = fobj_func
        return fobj

    def get_list(self, name):
        return eval(self.ids[name].text)

    def get_fobj_function(self):
        func_text = []
        function_full_text = self.ids.fobj.text.split('\n')
        simple_style = False
        ident = False
        ident_txt = ''
        replaces = {'y[]': 'y[:, 0]',
                    'x[': 'x[:, '}

        if 'use simple style' == function_full_text[0]:
            simple_style = True
            del function_full_text[0]
            func_text.append('    y = np.zeros((x.shape[0], 1))\n')
        for line in function_full_text:
            if simple_style:
                if line == '{':
                    ident = True
                    ident_txt += '    '
                    continue
                elif line == '}':
                    ident = False
                    ident_txt = ident_txt[4:]
                    continue
                if ident:
                    line = '%s%s' % (ident_txt, line)
                for expression in replaces.keys():
                    line = line.replace(expression, replaces[expression])
            func_text.append('    %s\n' %line)
        func_text.insert(0, '    import numpy as np\n')
        func_text.insert(0, 'def fobj(x):\n')
        func_text.append('    return y')
        func_text = """{}""".format(''.join(func_text))
        code = compile(func_text, '<string>', 'exec')
        exec_code = {}
        exec(code, exec_code)
        return exec_code['fobj']

    def on_checkbox_active(self, checkbox, value):
        with open('fobj_function.py', 'r') as f:
            fobj_text = f.readlines()
            self.ids.fobj.text = ''.join(fobj_text)
        if value:
            pass
        else:
            pass


class SearchGroupAlgorithmApp(App):
    pass


if __name__ == '__main__':
    from kivy.config import Config
    Config.set('graphics', 'width', '500')
    Config.set('graphics', 'height', '600')
    janela = SearchGroupAlgorithmApp()
    janela.run()



# coding: utf-8

import numpy as np


class ObjectiveFunc:

    def __init__(self, inf=[-10., -10.], sup=[10., 10.]):
        self.inf = np.array(inf)
        self.sup = np.array(sup)
        self.dim = self.inf.shape[0]
        self.range_5 = range(5)
        self.cos = np.cos
        self.zeros = np.zeros

    def evaluate(self, x):
        y = self.zeros((x.shape[0], 1))
        s1 = self.zeros(x.shape[0])
        s2 = self.zeros(x.shape[0])
        for i in self.range_5:
            n = i + 1
            s1 = s1 + n * self.cos((n + 1) * x[:, 0] + n)
            s2 = s2 + n * self.cos((n + 1) * x[:, 1] + n)
        y[:, 0] = s1 * s2
        return y


class ParamsSGA:
    def __init__(self, AlfaMin=0.01, AlfaInitial=2, NIterations=10000, GlobalIterationsRatio=0.3,
                 PopulationSize=100, SearchGroupRatio=0.1, NPerturbed=5, TournamentSize=4,
                 SmallValue=0.0025):
        self.AlfaMin = AlfaMin
        self.AlfaInitial = AlfaInitial
        self.NIterations = NIterations
        self.GlobalIterationsRatio = GlobalIterationsRatio
        self.PopulationSize = PopulationSize
        self.SearchGroupRatio = SearchGroupRatio
        self.NPerturbed = int(NPerturbed)
        self.PlotFamily = False
        self.TournamentSize = int(TournamentSize)
        self.elite = 1
        self.SmallValue = SmallValue
        self.alfa_local = False
        self.current_iteration = 0
        self.NIterationsGlobal = int(self.GlobalIterationsRatio*self.NIterations)
        self.NIterationsLocal = self.NIterations - self.NIterationsGlobal
        self.AlfaCorrectionMatrix = np.array([[1, -4 / (self.NIterationsGlobal)],
                                              [0.25, -1 / (4 * self.NIterationsGlobal)],
                                              [0, 0]])

    def update_iterarion(self):
        self.current_iteration += 1

    def update_AlfaCorrectionMatrix(self):
        self.AlfaCorrectionMatrix = np.array([[1, -4 / (self.NIterationsGlobal)],
                                         [0.25, -1 / (4 * self.NIterationsGlobal)],
                                         [0, 0]])

    def alfa(self, inf, sup):
        if self.alfa_local:
            _alfa = (self.NIterationsLocal - self.current_iteration)/(self.NIterationsLocal)
            _alfa *= self.AlfaMin
            _alfa += (self.SmallValue * self.AlfaMin)
        else:
            _alfa = (self.AlfaInitial * self.alfa_correction() + self.AlfaMin)
        return _alfa * (sup - inf)

    def alfa_correction(self):
        if self.current_iteration == 0:
            return 0
        v = self.AlfaCorrectionMatrix[:, 0] + self.AlfaCorrectionMatrix[:, 1] * self.current_iteration
        return v.max()


def configure():
    f = ObjectiveFunc()
    sga = ParamsSGA()
    return sga, f


if __name__ == '__main__':
    sga, fobj = configure()
    z = fobj.evaluate(np.array([[100*0.048579216311839, 100*0.054828050124931]]))
    print(z)



# coding: utf-8

import numpy as np
from pysga.family import Family
from pysga.configure import ParamsSGA, ObjectiveFunc
from pysga.tournament import Tournament


def leaders_indices(f, sga, tournament):
    indices = np.transpose(np.arange(0, f.search_group.size, 1))
    tournament_indices = tournament.tournament()
    tournament_indices = np.sort(tournament_indices)
    indices[np.arange(sga.elite, f.search_group.size)] = tournament_indices
    return indices


def run(sga=None, fobj=None):
    if not sga:
        sga = ParamsSGA()
    if not fobj:
        fobj = ObjectiveFunc()
    f = Family(sga, fobj)
    n_pertubed = int(f.search_group.size - sga.elite)
    tournament = Tournament(f.population.size, n_pertubed, sga.TournamentSize)
    rtournament = Tournament(f.search_group.size, sga.NPerturbed, sga.TournamentSize)
    f.update_leaders(leaders_indices(f, sga, tournament))
    iteration_data = np.zeros((sga.NIterations, f.dim + 1))

    for k in range(sga.NIterationsGlobal):
        perturbed_indices = rtournament.reverse_tournament()
        f.mutate_leaders(perturbed_indices)
        f.generation(sga.alfa(fobj.inf, fobj.sup))
        f.sort_leaders()
        iteration_data[k] = f.leaders[0]
        sga.update_iterarion()

    sga.alfa_local = True
    sga.current_iteration = 0

    for k in range(sga.NIterationsLocal):
        perturbed_indices = rtournament.reverse_tournament()
        f.mutate_leaders(perturbed_indices)
        f.local_generation(sga.alfa(fobj.inf, fobj.sup))
        f.sort_database()
        iteration_data[sga.NIterationsGlobal + k] = f.database[0]
        f.update_leaders(leaders_indices(f, sga, tournament))
        sga.update_iterarion()

    x = iteration_data[-1, 1:]
    y = iteration_data[-1, 0]
    return x, y


if __name__ == '__main__':
    x, y = run()
    print(x, y)



# coding: utf-8

import numpy as np


class Tournament:

    def __init__(self, total, n_pertubed, size):
        self.total = int(total)
        self.n_pertubed = int(n_pertubed)
        self.size = int(size)
        self.range_tournament = range(self.n_pertubed)
        self.tournament_size_array = np.ones((self.size, 1), dtype=np.int64) * total
        self.perturbed_indices = np.zeros(n_pertubed, dtype=np.int64)
        self.indices = np.arange(1, total)

    def tournament(self):
        indices = self.indices.copy()
        for i in self.range_tournament:
            c = self.tournament_size_array - np.random.randint(2, self.total + 1 - i, size=(self.size, 1))
            k = c.min()
            self.perturbed_indices[i] = indices[k]
            try:
                indices[i:k + 1] = indices[i - 1:k]
            except ValueError:
                pass
        return self.perturbed_indices

    def reverse_tournament(self):
        indices = self.indices.copy()
        for i in self.range_tournament:
            c = self.tournament_size_array - np.random.randint(2, self.total + 1 - i, size=(self.size, 1))
            k = c.max()
            self.perturbed_indices[i] = indices[k]
            try:
                indices[i:k + 1] = indices[i - 1:k]
            except ValueError:
                try:
                    indices[i:] = indices[i:k]
                except:
                    pass
        return self.perturbed_indices



# coding: utf-8
import numpy as np


# Example function
def fobj(x):
    y = np.zeros((x.shape[0], 1))
    s1 = np.zeros(x.shape[0])
    s2 = np.zeros(x.shape[0])
    for i in range(5):
        n = i + 1
        s1 = s1 + n * np.cos((n + 1) * x[:, 0] + n)
        s2 = s2 + n * np.cos((n + 1) * x[:, 1] + n)
    y[:, 0] = s1 * s2
    return y



#!/usr/bin/env python

import os
from setuptools import setup

__version__ = '1.2.1'


long_description = ''
if os.path.exists('README.md'):
    try:
        import pypandoc
        long_description = pypandoc.convert('README.md', 'rst')
    except(IOError, ImportError):
        long_description = open('README.md', 'r', encoding='utf8').read()


setup(name='pysga',
      version=__version__,
      description='Search Group Algorithm metaheuristic optimization method python adaptation',
      author='André Ginklings',
      author_email='andre.ginklings@gmail.com',
      url='https://github.com/Ginklings/pysga',
      keywords='metaheuristic optimization algorithm',
      classifiers=[
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Operating System :: OS Independent',
          'Topic :: Scientific/Engineering'],
      packages=['pysga'],
      package_data={'': ['searchgroupalgorithm.kv']},
      include_package_data=True,
      install_requires=['numpy'],
      extras_require={'full': ['docutils',
                               'pygments',
                               'pypiwin32',
                               'kivy_deps.sdl2==0.1.22',
                               'kivy_deps.glew==0.1.12',
                               'kivy_deps.gstreamer==0.1.17',
                               'kivy_deps.angle==0.1.9',
                               'kivy']}
      )



# coding: utf-8

import numpy as np

class Family:

    def __init__(self, sga, fobj):
        self.inf = fobj.inf
        self.sup = fobj.sup
        self.dim = fobj.dim
        self.fertility = fobj.evaluate
        self.population = self.Population(sga)
        self.search_group = self.SearchGroup(sga)
        self.family_size = self._define_groups_amounts()
        self.unit_family_size = []
        self.family_data_indices = []
        self.initial_family_cte()
        self.leaders = np.zeros([self.search_group.size, self.dim+1])
        self.database = None
        self.initial_database()
        self.family = []
        self.initial_family()
        self.xTemp = []
        self.initial_xtemp()
        self.pertubed_arange = np.repeat(np.arange(1, sga.NPerturbed+1, 1),
                                         self.dim).reshape(sga.NPerturbed, self.dim)

    def initial_family_cte(self):
        for i in self.search_group.range:
            self.unit_family_size.append(np.ones((self.family_size[i], 1)))
            indices = np.arange(0, self.family_size[i]+1) + self.partial_population_families(i)
            self.family_data_indices.append(indices)

    def initial_xtemp(self):
        for i in self.search_group.range:
            self.xTemp.append(np.zeros([self.family_size[i], self.dim]))

    def initial_family(self):
        self.family = []
        for i in self.search_group.range:
            self.family.append(np.zeros((self.family_size[i]+1, self.dim+1)))

    def initial_database(self):
        inf = np.tile(self.inf, (self.population.size, 1))
        sup = np.tile(self.sup, (self.population.size, 1))
        x = inf + (sup - inf) * np.random.random((self.population.size, self.dim))
        y = self.fertility(x)
        self.database = np.hstack((y, x))
        self.database = self.database[np.argsort(self.database[:, 0])]

    def update_leaders(self, indices):
        self.leaders = self.database[indices]

    def sort_leaders(self):
        self.leaders = self.leaders[np.argsort(self.leaders[:, 0])]

    def sort_database(self):
        self.database = self.database[np.argsort(self.database[:, 0])]

    def mutate_leaders(self, pertubed_indices):
        mutation = np.mean(self.leaders[:, 1:], axis=0) * np.ones((len(pertubed_indices), self.dim))
        mutation += np.std(self.leaders[:, 1:], axis=0, ddof=1) * \
                    self.pertubed_arange * (np.random.random((len(pertubed_indices), self.dim)) - 0.5)
        np.maximum(mutation, self.inf, out=mutation)
        np.minimum(mutation, self.sup, out=mutation)
        self.leaders[pertubed_indices, 1:] = mutation

    def _define_groups_amounts(self):
            w = (np.arange(0, 1 + 1. / self.search_group.size, 1. / self.search_group.size) ** 2) \
                * (self.population.size - self.search_group.size)
            v = np.zeros(self.search_group.size)
            for i in range(self.search_group.size - 1):
                v[self.search_group.size - 1 - i] = max(round(w[i + 1] - w[i]), 1)
            v[0] = self.population.size - self.search_group.size - v.sum()
            return np.vectorize(int)(v)

    def partial_population_families(self, n):
        return int(self.family_size[:n].sum() + n)

    def generation(self, alfa):
        for i in self.search_group.range:
            self.xTemp[i] = self.leaders[i][1: self.dim+1] * self.unit_family_size[i] + \
                                                       alfa * self.unit_family_size[i] * \
                                                       (np.random.random((self.family_size[i], self.dim)) - 0.5)
            np.maximum(self.xTemp[i], self.inf, out=self.xTemp[i])
            np.minimum(self.xTemp[i], self.sup, out=self.xTemp[i])
            values = self.fertility(self.xTemp[i])
            min_value_indice = values.argmin(axis=0)
            self.leaders[i] = np.hstack((values[min_value_indice], self.xTemp[i][min_value_indice, :]))

    def local_generation(self, alfa):
        for i in self.search_group.range:
            self.xTemp[i] = self.leaders[i][1: self.dim+1] * self.unit_family_size[i] + \
                                                       alfa * self.unit_family_size[i] * \
                                                       (np.random.random((self.family_size[i], self.dim)) - 0.5)
            np.maximum(self.xTemp[i], self.inf, out=self.xTemp[i])
            np.minimum(self.xTemp[i], self.sup, out=self.xTemp[i])
            self.family[i][0, :] = self.leaders[i, :]
            self.family[i][1:self.family_size[i] + 1, 1: self.dim + 1] = self.xTemp[i]
            self.family[i][1:, 0:1] = self.fertility(self.family[i][1:, 1: self.dim+1])
            self.database[self.family_data_indices[i], :] = self.family[i][:]

    class Population:

        def __init__(self, sga):
            self.size = sga.PopulationSize

    class SearchGroup:

        def __init__(self, sga):
            self.ratio = sga.SearchGroupRatio
            self.size = int(round(self.ratio*sga.PopulationSize))
            self.range = range(self.size)
